<?php
declare(strict_types=1);

namespace Inti\Wp\Provider\Plugin;

use DateTime;
use DateTimeZone;
use Inti\Wp\Plugin\PluginInterface;
use Inti\Wp\Provider\ObjectProviderInterface;
use Inti\Wp\Provider\Post\ProviderInterface;
use Inti\Wp\Provider\ProviderInterface as BaseProviderInterface;
use Inti\Wp\Provider\Service\IdOffsets;
use Inti\Wp\Provider\Service\ProviderCollection;
use Inti\Wp\Provider\Type\Post\PostTypeInterface;
use stdClass;
use WP_Query;

use function add_action;
use function add_filter;
use function array_filter;
use function array_map;
use function implode;
use function intval;
use function is_int;
use function is_string;
use function preg_match;
use function register_post_type;
use function wp_cache_set;

/**
 * Class ProvidePosts
 *
 * @package Inti\Wp\Provider\Plugin
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/oktan-nautika/inti-wp-provider
 */
final class ProvidePosts implements PluginInterface
{
    private ProviderCollection $providers;
    private IdOffsets $idOffsets;

    public function __construct(ProviderCollection $providers, IdOffsets $idOffsets)
    {
        $this->providers = $providers;
        $this->idOffsets = $idOffsets;
    }

    public function run(?array $options = null): void
    {
        $providers = $this->providers->getProviders(ProviderInterface::class);

        if (count($providers) === 0) {
            return;
        }

        add_action('init', [$this, 'register_post_types'], 0);
        add_filter('posts_pre_query', [$this, 'posts_pre_query'], 10, 2);
        add_filter('query', [$this, 'query']);
        add_filter('wp_count_posts', [$this, 'wp_count_posts'], 10, 2);

        foreach ($providers as $provider) {
            wp_cache_set('posts-' . $provider->getObjectType()->getKey(), new stdClass(), 'counts');
        }
    }

    public function posts_pre_query($posts, WP_Query $query): ?array
    {
        if (!is_string($query->query_vars['post_type']) || empty($query->query_vars['post_type'])) {
            return $posts;
        }

        if (
            ($provider = $this->providers->getProvider(ProviderInterface::class, $query->query_vars['post_type']))
            && $provider instanceof ObjectProviderInterface
        ) {
            $limit = $query->query_vars['posts_per_page'];
            $page = max(1, intval($query->query_vars['paged']));
            $offset = ($page - 1) * $limit;

            if (is_int($query->query_vars['p']) && $this->idOffsets->isOffset($query->query_vars['p'])) {
                $post_id = $this->idOffsets->unoffset($query->query_vars['p'], 'post');
                $rows = $provider->findByIds([$post_id]);
            } else if (is_array($query->query_vars['post__in']) && !empty($query->query_vars['post__in'])) {
                $post_ids = array_filter(array_map(function ($id) {
                    $id = intval($id);
                    return $this->idOffsets->isOffset($id) ? $this->idOffsets->unoffset($id, 'post') : false;
                }, $query->query_vars['post__in']));
                $rows = $provider->findByIds($post_ids);
            } else if (!empty($query->query_vars['s'])) {
                $rows = $provider->search($query->query_vars['s'], $limit, $offset);
            } else if (!empty($query->query_vars['name'])) {
                $rows = array_filter([
                    $provider->findBySlug($query->query_vars['name']),
                ]);
            } else {
                $rows = $provider->getAll($limit, $offset);
            }

            foreach ($rows as $row) {
                $posts[] = (object) $this->prepareRow($row, $provider);
            }

            $query->found_posts = $provider->count();
            $query->max_num_pages = ceil($query->found_posts / $limit);
        }

        return $posts;
    }

    public function query(string $query): string
    {
        if (preg_match('/^SELECT \* FROM wp_posts WHERE ID = (\d+) LIMIT 1$/', $query, $matches)) {
            $post_id = intval($matches[1]);

            if ($this->idOffsets->isOffset($post_id)) {
                $type = $this->idOffsets->getType($post_id, 'post');
                $post_id = $this->idOffsets->unoffset($post_id, 'post');

                if (
                    ($provider = $this->providers->getProvider(ProviderInterface::class, $type))
                    && $provider instanceof ObjectProviderInterface
                ) {
                    $rows = $provider->findByIds([$post_id]);

                    if (count($rows) > 0) {
                        $row = $this->prepareRow($rows[0], $provider);

                        foreach ($row as $field => $value) {
                            $fields[] = sprintf(
                                '%s as %s',
                                is_string($value) ? "'{$value}'" : $value,
                                $field
                            );
                        }

                        $query = 'SELECT ' . implode(',', $fields ?? []);
                    }
                }
            }
        }

        return $query;
    }

    public function register_post_types(): void
    {
        foreach ($this->providers->getProviders(ProviderInterface::class) as $provider) {
            $type = $provider->getObjectType();

            if ($type instanceof PostTypeInterface && !$type->isBuiltin()) {
                register_post_type($type->getKey(), [
                    'description' => $type->getDescription(),
                    'hierarchical' => $type->isHierarchical(),
                    'label' => $type->getLabel(),
                    'labels' => $type->getLabels(),
                    'public' => $type->isPublic(),
                    'publicly_queryable' => $type->isPubliclyQueryable(),
                    'show_ui' => $type->showUi(),
                    'show_in_menu' => $type->showInMenu(),
                    'show_in_nav_menus' => $type->showInNavMenus(),
                    'show_in_admin_bar' => $type->showInAdminBar(),
                    'menu_position' => $type->getMenuPosition(),
                    'menu_icon' => $type->getMenuIcon(),
                    'capability_type' => $type->getCapabilityType(),
                    'capabilities' => $type->getCapabilities(),
                    'map_meta_cap' => $type->mapMetaCap(),
                    'supports' => $type->getSupportedFeatures(),
                    'register_meta_box_cb' => $type->getMetaBoxCallback(),
                    'taxonomies' => $type->getTaxonomies(),
                    'has_archive' => $type->hasArchive(),
                    'rewrite' => $type->getRewriteConfig(),
                    'query_var' => $type->getQueryVar(),
                    'can_export' => false,
                    'delete_with_user' => false,
                    'show_in_rest' => false,
                    'rest_base' => false,
                    'rest_controller_class' => false,
                    'template' => [],
                    'template_lock' => false,
                ]);
            }
        }
    }

    public function wp_count_posts(object $count, string $type): object
    {
        if ($provider = $this->providers->getProviders(ProviderInterface::class, $type)[0]) {
            $count->publish = $provider->count();
        }

        return $count;
    }

    private function prepareRow(array $row, BaseProviderInterface $provider): array
    {
        $row['ID'] = $this->idOffsets->offset(intval($row['ID']), $provider);
        $row['post_author'] = 0;
        $row['post_date'] = $row['post_date'] ?? (new DateTime('now'))->format('Y-m-d H:i:s');
        /** @noinspection PhpUnhandledExceptionInspection */
        $row['post_date_gmt'] = $row['post_date_gmt'] ?? (new DateTime('now', new DateTimeZone('UTC')))->format('Y-m-d H:i:s');
        $row['post_content'] = $row['post_content'] ?? '';
        $row['post_title'] = $row['post_title'] ?? '';
        $row['post_excerpt'] = $row['post_excerpt'] ?? '';
        $row['post_status'] = $row['post_status'] ?? 'publish';
        $row['comment_status'] = 'closed';
        $row['ping_status'] = 'closed';
        $row['post_password'] = '';

        $row['post_type'] = $provider->getObjectType()->getKey();
        $row['post_mime_type'] = '';
        $row['comment_count'] = 0;

        return $row;
    }
}
